// THIS FILE IS GENERATED AUTOMATICALLY. DO NOT EDIT.
// LINT: LEGACY_NAMES

syntax = "proto3";

package vt.fileanalysis;

enum FileType {
  UNKNOWN = 0;
  PE_EXE = 1;
  PE_DLL = 2;
  MSI = 3;
  NE_EXE = 10;
  NE_DLL = 11;
  DOS_EXE = 20;
  DOS_COM = 21;
  COFF = 30;
  ELF = 31;
  LINUX_KERNEL = 32;
  RPM = 33;
  LINUX = 34;
  MACH_O = 35;
  JAVA_BYTECODE = 36;
  DMG = 37;
  DEB = 38;
  PKG = 39;
  PYC = 40;
  LNK = 50;
  DESKTOP_ENTRY = 51;
  JPEG = 100;
  TIFF = 101;
  GIF = 102;
  PNG = 103;
  BMP = 104;
  GIMP = 105;
  IN_DESIGN = 106;
  PSD = 107;
  TARGA = 108;
  XWD = 109;
  DIB = 110;
  JNG = 111;
  ICO = 112;
  FPX = 113;
  EPS = 114;
  SVG = 115;
  EMF = 116;
  WEBP = 117;
  DWG = 118;
  DXF = 119;
  UNUSED_THREEDS = 120;
  OGG = 150;
  FLC = 151;
  FLI = 152;
  MP3 = 153;
  FLAC = 154;
  WAV = 155;
  MIDI = 156;
  AVI = 157;
  MPEG = 158;
  QUICKTIME = 159;
  ASF = 160;
  DIVX = 161;
  FLV = 162;
  WMA = 163;
  WMV = 164;
  RM = 165;
  MOV = 166;
  MP4 = 167;
  T3GP = 168;
  WEBM = 169;
  MKV = 170;
  PDF = 200;
  PS = 201;
  DOC = 202;
  DOCX = 203;
  PPT = 204;
  PPTX = 205;
  PPSX = 209;
  XLS = 206;
  XLSX = 207;
  RTF = 208;
  ODP = 250;
  ODS = 251;
  ODT = 252;
  HWP = 253;
  GUL = 254;
  ODF = 255;
  ODG = 256;
  ONE_NOTE = 257;
  UNUSED_OOXML = 258;
  SLK = 259;
  EBOOK = 260;
  LATEX = 261;
  TTF = 262;
  EOT = 263;
  WOFF = 264;
  CHM = 265;
  ZIP = 300;
  GZIP = 301;
  BZIP = 302;
  RZIP = 303;
  DZIP = 304;
  SEVENZIP = 305;
  CAB = 306;
  JAR = 307;
  RAR = 308;
  MSCOMPRESS = 309;
  ACE = 310;
  ARC = 311;
  ARJ = 312;
  ASD = 313;
  BLACKHOLE = 314;
  KGB = 315;
  ZLIB = 316;
  TAR = 317;
  ZST = 318;
  LZFSE = 319;
  PYTHON_WHL = 320;
  PYTHON_PKG = 321;
  MSIX = 322;
  TEXT = 400;
  SCRIPT = 401;
  PHP = 402;
  PYTHON = 403;
  PERL = 404;
  RUBY = 405;
  C = 406;
  CPP = 407;
  JAVA = 408;
  SHELLSCRIPT = 409;
  PASCAL = 410;
  AWK = 411;
  DYALOG = 412;
  FORTRAN = 413;
  JAVASCRIPT = 414;
  POWERSHELL = 415;
  VBA = 416;
  M4 = 417;
  OBJETIVEC = 418;
  JMOD = 419;
  MAKEFILE = 420;
  INI = 421;
  UNUSED_CLJ = 422;
  UNUSED_PDB = 425;
  SQL = 426;
  UNUSED_NEKO = 427;
  UNUSED_WER = 428;
  GOLANG = 429;
  M3U = 430;
  BAT = 431;
  MSC = 432;
  RDP = 433;
  SYMBIAN = 500;
  PALMOS = 501;
  WINCE = 502;
  ANDROID = 503;
  IPHONE = 504;
  HTML = 600;
  XML = 601;
  SWF = 602;
  FLA = 603;
  COOKIE = 604;
  TORRENT = 605;
  EMAIL = 606;
  OUTLOOK = 607;
  SGML = 608;
  JSON = 609;
  CSV = 610;
  HTA = 611;
  INTERNET_SHORTCUT = 612;
  CAP = 700;
  ISOIMAGE = 800;
  SQUASHFS = 801;
  VHD = 802;
  VHDX = 803;
  APPLE = 1000;
  MACINTOSH = 1001;
  APPLESINGLE = 1002;
  APPLEDOUBLE = 1003;
  MACINTOSH_HFS = 1004;
  APPLE_PLIST = 1005;
  MACINTOSH_LIB = 1006;
  APPLESCRIPT = 1007;
  APPLESCRIPT_COMPILED = 1008;
  CRX = 1100;
  XPI = 1101;
  ROM = 1200;
  IPS = 1201;
  PEM = 1300;
  PGP = 1301;
  CRT = 1302;
}
